/*
 * Copyright (c) 2024. Das Land Schleswig-Holstein
 * vertreten durch den Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */

package de.ozgcloud.nachrichten.antragraum;

import java.io.IOException;
import java.io.InputStream;
import java.security.cert.CertificateException;
import java.security.cert.CertificateFactory;
import java.security.cert.X509Certificate;
import java.security.interfaces.RSAPrivateKey;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.Set;

import org.opensaml.core.config.ConfigurationService;
import org.opensaml.core.config.InitializationService;
import org.opensaml.core.criterion.EntityIdCriterion;
import org.opensaml.core.xml.XMLObject;
import org.opensaml.core.xml.config.XMLObjectProviderRegistry;
import org.opensaml.saml.common.xml.SAMLConstants;
import org.opensaml.saml.criterion.ProtocolCriterion;
import org.opensaml.saml.metadata.criteria.role.impl.EvaluableProtocolRoleDescriptorCriterion;
import org.opensaml.saml.saml2.metadata.EntitiesDescriptor;
import org.opensaml.saml.saml2.metadata.EntityDescriptor;
import org.opensaml.saml.saml2.metadata.KeyDescriptor;
import org.opensaml.security.credential.Credential;
import org.opensaml.security.credential.CredentialResolver;
import org.opensaml.security.credential.UsageType;
import org.opensaml.security.credential.criteria.impl.EvaluableEntityIDCredentialCriterion;
import org.opensaml.security.credential.criteria.impl.EvaluableUsageCredentialCriterion;
import org.opensaml.security.credential.impl.CollectionCredentialResolver;
import org.opensaml.security.criteria.UsageCriterion;
import org.opensaml.security.x509.BasicX509Credential;
import org.opensaml.xmlsec.config.impl.DefaultSecurityConfigurationBootstrap;
import org.opensaml.xmlsec.keyinfo.KeyInfoSupport;
import org.opensaml.xmlsec.signature.support.SignatureTrustEngine;
import org.opensaml.xmlsec.signature.support.impl.ExplicitKeySignatureTrustEngine;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.autoconfigure.condition.ConditionalOnProperty;
import org.springframework.context.annotation.Configuration;
import org.springframework.core.io.Resource;
import org.springframework.security.converter.RsaKeyConverters;
import org.springframework.security.saml2.Saml2Exception;
import org.springframework.security.saml2.core.Saml2X509Credential;
import org.springframework.util.Assert;

import jakarta.annotation.PostConstruct;
import lombok.Getter;
import net.shibboleth.utilities.java.support.component.ComponentInitializationException;
import net.shibboleth.utilities.java.support.resolver.CriteriaSet;
import net.shibboleth.utilities.java.support.xml.BasicParserPool;
import net.shibboleth.utilities.java.support.xml.ParserPool;
import net.shibboleth.utilities.java.support.xml.XMLParserException;

@Configuration
@ConditionalOnProperty(AntragraumProperties.PROPERTY_ANTRAGSRAUM_URL)
class BayernIdSamlConfiguration {
	private XMLObjectProviderRegistry registry;
	@Getter
	private ParserPool parserPool; // TODO als spring bean bereit stellen
	@Autowired
	private AntragraumProperties antragraumProperties;

	@PostConstruct
	void initOpenSAML() {
		try {
			registry = new XMLObjectProviderRegistry();
			ConfigurationService.register(XMLObjectProviderRegistry.class, registry);
			parserPool = initParserPool();

			registry.setParserPool(parserPool);
			InitializationService.initialize();
		} catch (Exception e) {
			throw new RuntimeException("Initialization failed");
		}
	}

	private ParserPool initParserPool() throws ComponentInitializationException {
		var localParserPool = new BasicParserPool();

		final var features = createFeatureMap();

		localParserPool.setBuilderFeatures(features);

		localParserPool.setBuilderAttributes(new HashMap<>());

		localParserPool.initialize();

		return localParserPool;
	}

	private Map<String, Boolean> createFeatureMap() {
		final Map<String, Boolean> features = new HashMap<>();
		features.put("http://xml.org/sax/features/external-general-entities", Boolean.FALSE);
		features.put("http://xml.org/sax/features/external-parameter-entities", Boolean.FALSE);
		features.put("http://apache.org/xml/features/disallow-doctype-decl", Boolean.TRUE);
		features.put("http://apache.org/xml/features/validation/schema/normalized-value", Boolean.FALSE);
		features.put("http://javax.xml.XMLConstants/feature/secure-processing", Boolean.TRUE);
		return features;
	}

	SignatureTrustEngine getTrustEngine() {
		Set<Credential> credentials = new HashSet<>();
		Collection<Saml2X509Credential> keys = getCertificatesFromMetadata();

		for (Saml2X509Credential key : keys) {
			var cred = new BasicX509Credential(key.getCertificate());
			cred.setUsageType(UsageType.SIGNING);
			cred.setEntityId(antragraumProperties.getEntityId());
			credentials.add(cred);
		}

		CredentialResolver credentialsResolver = new CollectionCredentialResolver(credentials);
		return new ExplicitKeySignatureTrustEngine(credentialsResolver,
				DefaultSecurityConfigurationBootstrap.buildBasicInlineKeyInfoCredentialResolver());
	}

	CriteriaSet getVerificationCriteria() {
		var criteria = new CriteriaSet();
		criteria.add(new EvaluableEntityIDCredentialCriterion(new EntityIdCriterion(antragraumProperties.getEntityId())));
		criteria.add(new EvaluableProtocolRoleDescriptorCriterion(new ProtocolCriterion("urn:oasis:names:tc:SAML:2.0:protocol")));
		criteria.add(new EvaluableUsageCredentialCriterion(new UsageCriterion(UsageType.SIGNING)));
		return criteria;
	}

	Saml2X509Credential getDecryptionCredential() {
		var privateKey = readPrivateKey(antragraumProperties.getDecryptionPrivateKey());
		var certificate = readCertificateFromResource(antragraumProperties.getDecryptionCertificate());
		return new Saml2X509Credential(privateKey, certificate, Saml2X509Credential.Saml2X509CredentialType.DECRYPTION);
	}

	private RSAPrivateKey readPrivateKey(Resource location) {
		Assert.state(location != null, "No private key location specified");
		Assert.state(location.exists(), () -> "Private key location '" + location + "' does not exist");
		try (var inputStream = location.getInputStream()) {
			return RsaKeyConverters.pkcs8().convert(inputStream);
		} catch (IOException e) {
			throw new IllegalArgumentException(e);
		}
	}

	private X509Certificate readCertificateFromResource(Resource location) {
		Assert.state(location != null, "No certificate location specified");
		Assert.state(location.exists(), () -> "Certificate  location '" + location + "' does not exist");
		try (var inputStream = location.getInputStream()) {

			return (X509Certificate) CertificateFactory.getInstance("X.509").generateCertificate(inputStream);
		} catch (IOException | CertificateException e) {
			throw new IllegalArgumentException(e);
		}
	}

	List<Saml2X509Credential> getCertificatesFromMetadata() {
		try (var metadata = antragraumProperties.getMetadataUri().getInputStream()) {
			var xmlObject = xmlObject(metadata);
			var descriptorOptional = findEntityDescriptor(xmlObject);
			if (descriptorOptional.isPresent()) {
				return getVerificationCertificates(descriptorOptional.get());
			}
		} catch (IOException e) {
			throw new Saml2Exception("Error reading idp metadata.", e);
		} catch (ComponentInitializationException | XMLParserException e) {
			throw new Saml2Exception("Error initializing parser pool.", e);
		}

		throw new Saml2Exception("No IDPSSO Descriptors found");
	}

	private XMLObject xmlObject(InputStream inputStream) throws ComponentInitializationException, XMLParserException {
		var document = getParserPool().parse(inputStream);
		var element = document.getDocumentElement();
		var unmarshaller = this.registry.getUnmarshallerFactory().getUnmarshaller(element);
		if (unmarshaller == null) {
			throw new Saml2Exception("Unsupported element of type " + element.getTagName());
		}
		try {
			return unmarshaller.unmarshall(element);
		} catch (Exception ex) {
			throw new Saml2Exception(ex);
		}
	}

	private Optional<EntityDescriptor> findEntityDescriptor(XMLObject metadata) {
		Optional<EntityDescriptor> descriptor = Optional.empty();
		if (metadata instanceof EntityDescriptor entityDescritor) {
			descriptor = Optional.of(entityDescritor);
		} else if (metadata instanceof EntitiesDescriptor entitiesDescriptor) {
			descriptor = entitiesDescriptor.getEntityDescriptors().stream().findFirst();
		}

		return descriptor.filter(entityDescriptor -> entityDescriptor.getIDPSSODescriptor(SAMLConstants.SAML20P_NS) != null).stream().findFirst();
	}

	List<Saml2X509Credential> getVerificationCertificates(EntityDescriptor descriptor) {
		var idpssoDescriptor = descriptor.getIDPSSODescriptor(SAMLConstants.SAML20P_NS);
		if (idpssoDescriptor == null) {
			throw new Saml2Exception("Metadata response is missing the necessary IDPSSODescriptor element");
		}
		List<Saml2X509Credential> verification = new ArrayList<>();
		for (KeyDescriptor keyDescriptor : idpssoDescriptor.getKeyDescriptors()) {
			if (keyDescriptor.getUse().equals(UsageType.SIGNING)) {
				var certificates = certificates(keyDescriptor);
				for (X509Certificate certificate : certificates) {
					verification.add(Saml2X509Credential.verification(certificate));
				}
			}
		}
		if (verification.isEmpty()) {
			throw new Saml2Exception(
					"Metadata response is missing verification certificates, necessary for verifying SAML assertions");
		}

		return verification;
	}

	private List<X509Certificate> certificates(KeyDescriptor keyDescriptor) {
		try {
			return KeyInfoSupport.getCertificates(keyDescriptor.getKeyInfo());
		} catch (CertificateException ex) {
			throw new Saml2Exception(ex);
		}
	}
}
